#  SPDX-License-Identifier: Apache-2.0
"""
Python Package for controlling Porsche Connect API.

"""
import asyncio
import calendar
import datetime
import json
import logging
import time
import base64
import hashlib
import os
import re
import urllib.parse
from typing import Dict, Text
from bs4 import BeautifulSoup

import aiohttp

try:
    from rich import print
except ImportError:
    pass

from .exceptions import WrongCredentials, PorscheException

_LOGGER = logging.getLogger(__name__)


async def on_request_start(session, trace_config_ctx, params):
    _LOGGER.debug("Starting request")
    _LOGGER.debug(params)


async def on_request_end(session, trace_config_ctx, params):
    _LOGGER.debug("Ending request")
    _LOGGER.debug(params)


trace_config = aiohttp.TraceConfig()
trace_config.on_request_start.append(on_request_start)
trace_config.on_request_end.append(on_request_end)

AUTHORIZATION_SERVER="identity.porsche.com"
REDIRECT_URI="https://my.porsche.com/"
AUDIENCE="https://api.porsche.com"
TENANT="porsche-production"
COUNTRY="de"
LANGUAGE="de_DE"
CLIENT_ID="UYsK00My6bCqJdbQhTQ0PbWmcSdIAMig"



class Connection:
    """Connection to Porsche Connect API."""

    def __init__(
        self,
        email: Text = None,
        password: Text = None,
        websession: aiohttp.ClientSession = None,
        language: Text = "de",
        country: Text = "DE",
        tokens=None,
    ) -> None:
        """Initialize connection object."""
        self.porscheApplications = {
            "api": {
                "client_id": CLIENT_ID,
                "redirect_uri": REDIRECT_URI,
                "prefix": "https://api.porsche.com/core/api/",
            },
            "profile": {
                "client_id": CLIENT_ID,
                "api_key": "QPw3VOLAMfI7r0nmRY8ELq4RzZpZeXEE",
                "redirect_uri": REDIRECT_URI,
                "prefix": "https://api.porsche.com/profiles",
            },
            # "auth": {
            #     "client_id": "4mPO3OE5Srjb1iaUGWsbqKBvvesya8oA",
            #     "redirect_uri": "https://my.porsche.com/core/de/de_DE/",
            #     "prefix": "https://login.porsche.com",
            # },
            "carcontrol": {
                #"client_id": "Ux8WmyzsOAGGmvmWnW7GLEjIILHEztAs",
                "client_id": CLIENT_ID,
                "redirect_uri": "https://my.porsche.com/myservices/auth/auth.html",
                "prefix": "https://api.porsche.com/",
            },
        }

        self.isTokenRefreshed = False
        self.tokens = tokens or {}
        self.email = email
        self.password = password
        self.websession = websession
        self._isLoggedIn = False
        self.country = country
        self.language = language
        self.auth_state = {}

        if self.websession is None:
            self.websession = aiohttp.ClientSession(trace_configs=[trace_config])
        _LOGGER.debug("New connection prepared")

    async def _login(self):
        _LOGGER.debug("Start authentication, get initial state from auth server")
        # Do not follow redirect
        start_login_url = f"https://{AUTHORIZATION_SERVER}/authorize?response_type=code&client_id={CLIENT_ID}&code_challenge_method=S256&redirect_uri={REDIRECT_URI}&ui_locales=de-DE&audience={AUDIENCE}&scope=openid"
        async with self.websession.get(start_login_url, allow_redirects=False) as resp:
            location = resp.headers["Location"]
            params = urllib.parse.parse_qs(urllib.parse.urlparse(location).query)
            _LOGGER.debug(params)
            have_code = params.get('code', None)
            if have_code is not None:
                _LOGGER("We already have a code in session, skip login")
                self.auth_state['code'] = have_code
                return
            self.auth_state["state"] = params["state"][0]
            self.auth_state["client"] = params["client"][0]
        _LOGGER.debug(self.auth_state)


        # Post auth data
        _LOGGER.debug("POST authentication details....")
        auth_body = {
                "sec": "high",
                "username": self.email,
                "password": self.password,
                "code_challenge_method": "S256",
                "redirect_uri": REDIRECT_URI,
                "ui_locales": "de-DE",
                "audience": AUDIENCE,
                "client_id": CLIENT_ID,
                "connection": "Username-Password-Authentication",
                "state": self.auth_state["state"],
                "tenant": TENANT,
                "response_type": "code"
                }
        auth_url = f"https://{AUTHORIZATION_SERVER}/usernamepassword/login"
        verify_body = {} 
        async with self.websession.post(auth_url, headers={"Content-Type": "application/x-www-form-urlencoded"}, data=auth_body, max_redirects=30) as resp:
            # In case of wrong credentials there is a state param in the redirect url

            if resp.status == 401:
                message = await resp.json()
                raise WrongCredentials(message['message'])

            html_body = await resp.text()

            _LOGGER.debug(resp.status)
            _LOGGER.debug(html_body)

            soup = BeautifulSoup(html_body,features="html.parser")
            hidden_tags = soup.find_all("input", type="hidden")
            for tag in hidden_tags:
                verify_body[tag.attrs['name']] = tag.attrs['value']
            _LOGGER.debug(verify_body)

        # Follow callback
        _LOGGER.debug("POST authentication verification...")
        auth_url = f"https://{AUTHORIZATION_SERVER}/login/callback"
        resume_url = ""
        async with self.websession.post(auth_url, headers={"Content-Type": "application/x-www-form-urlencoded"}, data=verify_body, allow_redirects=False) as resp:
            resume_url = resp.headers['Location']
            _LOGGER.debug(f"Resume at {resume_url}")

        _LOGGER.debug("Sleeping 2.5s...")
        await asyncio.sleep(2.5)

        # Resume auth
        auth_url = f"https://{{AUTHORIZATION_SERVER}}{resume_url}"
        async with self.websession.get(start_login_url, allow_redirects=False) as resp:
            location = resp.headers["Location"]
            params = urllib.parse.parse_qs(urllib.parse.urlparse(location).query)
            _LOGGER.debug(params)
            self.auth_state["code"] = params["code"][0]
            _LOGGER.debug(f"Got code {self.auth_state['code']}")

        _LOGGER.debug("Sleeping 2.5s...")
        await asyncio.sleep(2.5)
        
        self._isLoggedIn = True
        return True

    async def getAllTokens(self):
        now = calendar.timegm(datetime.datetime.now().timetuple())
        for applicationKey in self.porscheApplications:
            application = self.porscheApplications[applicationKey]
            _LOGGER.debug(f"Get token for app {applicationKey}")
            token = self.tokens.get(application["client_id"], None)
            if token is None or token["expiration"] < now:
                token = await self._requestToken(application)
                self.tokens[application["client_id"]] = token
        self.isTokenRefreshed = False
        return self.tokens

    async def _requestToken(self, application: Dict, wasExpired=False):
        if not self._isLoggedIn or wasExpired:
            await self._login()


        _LOGGER.debug("POST to acces token endpoint...")
        auth_url = f"https://{AUTHORIZATION_SERVER}/oauth/token"
        auth_body = {
                "client_id": application['client_id'],
                "grant_type": "authorization_code",
                "code": self.auth_state['code'],
                "redirect_uri": REDIRECT_URI
                }
        _LOGGER.debug(auth_body)
        _LOGGER.debug( "Requesting access token for client id %s", application["client_id"])
        now = calendar.timegm(datetime.datetime.now().timetuple())
        async with self.websession.post(auth_url, headers={"Content-Type": "application/x-www-form-urlencoded"}, data=auth_body, max_redirects=30) as resp:
            _LOGGER.debug(f"Response status {resp.status}")
            token_data = await resp.json()
            _LOGGER.debug(token_data)
            jwt = self.jwt_payload_decode(token_data["access_token"])
            token = token_data
            token["expiration"] = now + token_data.get("exp", 3600)
            token["decoded_token"] = jwt
            token["apikey"] = jwt["azp"]
            _LOGGER.debug("Token: %s", token)
            self.isTokenRefreshed = True
            return token


    async def get(self, url, params=None):
        try:
            application = self._applicationForURL(url)
            headers = await self._createhead(application)
            async with self.websession.get(url, params=params, headers=headers) as resp:
                return await resp.json()
        except aiohttp.ClientResponseError as exception_:
            raise PorscheException(exception_.status)

    async def post(self, url, data=None, json=None):
        try:
            application = self._applicationForURL(url)
            headers = await self._createhead(application)
            async with self.websession.post(
                url, data=data, json=json, headers=headers
            ) as resp:
                return await resp.json()
        except aiohttp.ClientResponseError as exception_:
            raise PorscheException(exception_.status)

    async def put(self, url, data=None, json=None):
        try:
            application = self._applicationForURL(url)
            headers = await self._createhead(application)
            async with self.websession.put(
                url, data=data, json=json, headers=headers
            ) as resp:
                return await resp.json()
        except aiohttp.ClientResponseError as exception_:
            raise PorscheException(exception_.status)

    async def delete(self, url, data=None, json=None):
        try:
            application = self._applicationForURL(url)
            headers = await self._createhead(application)
            async with self.websession.delete(
                url, data=data, json=json, headers=headers
            ) as resp:
                return await resp.json()
        except aiohttp.ClientResponseError as exception_:
            raise PorscheException(exception_.status)

    def _applicationForURL(self, url):
        for key in self.porscheApplications:
            app = self.porscheApplications[key]
            if url.startswith(app["prefix"]):
                return app
        # else return None
        return None

    async def _createhead(self, application):
        # If no application matched the request URL then no auth headers are added
        if application is None:
            return {}
        now = calendar.timegm(datetime.datetime.now().timetuple())
        token = self.tokens.get(application["client_id"], None)
        if token is None or token["expiration"] < now:
            token = await self._requestToken(
                application, wasExpired=(token is not None)
            )
            self.tokens[application["client_id"]] = token
        head = {
            "Authorization": f"Bearer {token['access_token']}",
            "origin": "https://my.porsche.com",
            "apikey": application.get("api_key", token["apikey"]),
            "x-vrs-url-country": self.country.lower(),
            "x-vrs-url-language": f"{self.language.lower()}_{self.country.upper()}",
        }
        return head

    def _b64_decode(self, data):
        data += "=" * (4 - len(data) % 4)
        return base64.b64decode(data).decode("utf-8")

    def jwt_payload_decode(self, jwt):
        _, payload, _ = jwt.split(".")
        return json.loads(self._b64_decode(payload))

    async def close(self):
        await self.websession.close()



from pyporscheconnectapi.connection import Connection
from pyporscheconnectapi.exceptions import WrongCredentials
from hashlib import sha512

import asyncio
import datetime
import logging

_LOGGER = logging.getLogger(__name__)


class Client:
    """Client for Porsche Connect API."""

    def __init__(
        self,
        connection: Connection,
        country="de",
        language="de",
        timezone="Europe/Stockholm",
    ) -> None:
        self._connection = connection
        self.country = country
        self.language = language
        self.timezone = timezone
        self.locale_str = f"{country.lower()}/{language.lower()}_{country.upper()}"

    def isTokenRefreshed(self):
        return self._connection.isTokenRefreshed

    async def getAllTokens(self):
        return await self._connection.getAllTokens()

    async def _spinner(self, url):
        for iteration in range(120):
            data = await self._connection.get(url)
            if "status" in data and data["status"] != "IN_PROGRESS":
                break
            elif "actionState" in data and data["actionState"] != "IN_PROGRESS":
                break
            await asyncio.sleep(1)
        return data

    async def _lockUnlock(self, vin, pin, action, waitForConfirmation=True):
        if pin is not None:
            challengeResult = await self._connection.get(
                f"https://api.porsche.com/service-vehicle/remote-lock-unlock/{vin}/{action}"
            )

            challenge = challengeResult.get("challenge")
            token = challengeResult.get("securityToken")
            pinhash = sha512(bytes.fromhex(pin+challenge)).hexdigest().upper()
            payload = {"challenge": challenge, "securityPinHash": pinhash, "securityToken": token}
        else:
            payload = {}

        progressResult = await self._connection.post(
            f"https://api.porsche.com/service-vehicle/remote-lock-unlock/{vin}/{action}",
            json=payload,
        )

        error = progressResult.get("pcckErrorKey", None)
        if error == "INCORRECT":
            raise WrongCredentials("PIN code was incorrect")
        elif error == "LOCKED_60_MINUTES":
            raise WrongCredentials("Too many failed attempts, locked 60 minutes")
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/service-vehicle/remote-lock-unlock/{vin}/{progressResult['requestId']}/status"
        )
        if result["status"] == "SUCCESS":
            result = await self._connection.get(
                f"https://api.porsche.com/service-vehicle/remote-lock-unlock/{vin}/last-actions"
            )
        return result

    async def _setClimate(self, vin, action, waitForConfirmation=True):
        progressResult = await self._connection.post(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{vin}/toggle-direct-climatisation/{action}",
            json={},
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{vin}/toggle-direct-climatisation/status/{progressResult['requestId']}"
        )
        return result

    async def _setDirectCharge(
        self,
        vin,
        action,
        model=None,
        waitForConfirmation=True,
    ):
        if model is None:
            data = await self.getCapabilities(vin)
            model = data["carModel"]
        progressResult = await self._connection.post(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/toggle-direct-charging/{action}",
            json={},
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/toggle-direct-charging/status/{progressResult['requestId']}"
        )
        return result

    async def _setHonkAndFlash(self, vin, waitForConfirmation=True):
        progressResult = await self._connection.post(
            f"https://api.porsche.com/service-vehicle/honk-and-flash/{vin}/honk-and-flash",
            json={},
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/service-vehicle/honk-and-flash/{vin}/{progressResult['id']}/status"
        )
        return result

    async def _setFlash(self, vin, waitForConfirmation=True):
        progressResult = await self._connection.post(
            f"https://api.porsche.com/service-vehicle/honk-and-flash/{vin}/flash",
            json={},
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/service-vehicle/honk-and-flash/{vin}/{progressResult['id']}/status"
        )
        return result

    async def _addTimer(self, vin, timer, model=None, waitForConfirmation=True):
        """Add new charge & climate timer"""
        if model is None:
            data = await self.getCapabilities(vin)
            model = data["carModel"]
        progressResult = await self._connection.post(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/timer",
            json=timer,
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/action-status/{progressResult['actionId']}"
        )
        return result

    async def _updateChargingProfile(
        self,
        vin,
        model=None,
        profile=None,
        waitForConfirmation=True,
    ):
        if model is None:
            data = await self.getCapabilities(vin)
            model = data["carModel"]

        progressResult = await self._connection.put(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/profile",
            json=profile,
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/action-status/{progressResult['actionId']}"
        )
        return result

    async def _updateTimer(
        self,
        vin,
        timer,
        timerID="1",
        model=None,
        waitForConfirmation=True,
    ):
        if model is None:
            data = await self.getCapabilities(vin)
            model = data["carModel"]
        """Update existing charge & climate timer"""
        timer.update({"timerID": timerID})
        progressResult = await self._connection.put(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/timer",
            json=timer,
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/action-status/{progressResult['actionId']}"
        )
        return result

    async def _deleteTimer(
        self, vin, timerID="1", model=None, waitForConfirmation=True
    ):
        if model is None:
            data = await self.getCapabilities(vin)
            model = data["carModel"]
        """Delete existing charge & climate timer"""
        progressResult = await self._connection.delete(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/timer/{timerID}"
        )
        if not waitForConfirmation:
            return progressResult
        result = await self._spinner(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}/action-status/{progressResult['actionId']}"
        )
        return result

    def _formatTimer(self, active, charge_settings, climate, time_date):
        """Format combined payload for charge & climate timer"""
        payload = {"active": active, "climatised": climate}
        payload.update(charge_settings)
        payload.update(time_date)
        return payload

    def _formatChargeTimer(self, charge=False, target_charge=10):
        """Format payload for charge timer
        Parameters:
        charge (bool): Enable charging?
        target_charge (0-100): Target charge level
        """
        charge_level = min(max(int(target_charge), 0), 100)
        return {"chargeOption": charge, "targetChargeLevel": charge_level}

    def _formatTimerTime(self, time, repeat_days=[]):
        """Format time/date settings for timer
        Parameters:
        time (datetime): Target departure date/time as datetime object
        repeat_days (list of integers): Which days to repeat on, if any, 0=Monday)
        """

        # FIXME: There's something missing here for creation of repeating timers
        # Hard to debug b/c the website also isn't letting me right now (app is fine)
        # Format seems to work for updating timers
        if len(repeat_days):
            frequency = "CYCLIC"
        else:
            frequency = "SINGLE"
        repeat_dict = {
            "MONDAY": False,
            "TUESDAY": False,
            "WEDNESDAY": False,
            "THURSDAY": False,
            "FRIDAY": False,
            "SATURDAY": False,
            "SUNDAY": False,
        }
        for i in repeat_days:
            if i == 0:
                repeat_dict["MONDAY"] = True
            elif i == 1:
                repeat_dict["TUESDAY"] = True
            elif i == 2:
                repeat_dict["WEDNESDAY"] = True
            elif i == 3:
                repeat_dict["THURSDAY"] = True
            elif i == 4:
                repeat_dict["FRIDAY"] = True
            elif i == 5:
                repeat_dict["SATURDAY"] = True
            elif i == 6:
                repeat_dict["SUNDAY"] = True

        return {
            "departureDateTime": time.isoformat(),
            "frequency": frequency,
            "weekDays": repeat_dict,
        }

    async def newTimer(
        self,
        vin,
        time,
        active=False,
        charge=False,
        target_charge=10,
        climate=False,
        repeat_days=[],
        model=None,
        waitForConfirmation=True,
    ):
        """Create a new timer on the vehicle
        Parameters:
        vin (string): Vehicle VIN
        time (datetime): Target departure date/time as datetime object
        active (bool): Timer should be active upon creation?
        charge (bool): Enable charging?
        target_charge (0-100): Target charge level
        climate (bool): Precondition cabin?
        repeat_days (list of integers): Which days to repeat on, if any, 0=Monday)

        FIXME: Creation of repeating timers isn't working, see _formatTimerTime
        """
        timer_time = self._formatTimerTime(time, repeat_days)
        charge_timer = self._formatChargeTimer(charge, target_charge)
        full_timer = self._formatTimer(active, charge_timer, climate, timer_time)
        return await self._addTimer(
            vin, full_timer, model, waitForConfirmation=waitForConfirmation
        )

    async def updateChargingProfile(
        self,
        vin,
        model,
        profileId: int,
        minimumChargeLevel: int = None,
        long: float = None,
        lat: float = None,
        profileActive: bool = None,
        waitForConfirmation=True,
    ):
        emobility = (await self.getEmobility(vin, model=model))["chargingProfiles"][
            "profiles"
        ]
        profile = {item["profileId"]: item for item in emobility}[profileId]

        if minimumChargeLevel is not None:
            minimumChargeLevel = min(max(int(minimumChargeLevel), 25), 100)
            profile["chargingOptions"]["minimumChargeLevel"] = minimumChargeLevel

        if profileActive is not None:
            profile["profileActive"] = profileActive

        if long is not None:
            profile["position"]["longitude"] = long

        if lat is not None:
            profile["position"]["latitude"] = lat

        return await self._updateChargingProfile(
            vin, model, profile, waitForConfirmation=waitForConfirmation
        )

    async def updateTimer(
        self,
        vin,
        time,
        timerID="1",
        active=False,
        charge=False,
        target_charge=10,
        climate=False,
        repeat_days=[],
        model=None,
        waitForConfirmation=True,
    ):
        """Update an existing timer on the vehicle
        Parameters:
        vin (string): Vehicle VIN
        time (datetime): Target departure date/time as datetime object
        timerId (string): Target timer (numeric string)
        active (bool): Timer should be active upon creation?
        charge (bool): Enable charging?
        target_charge (0-100): Target charge level
        climate (bool): Precondition cabin?
        repeat_days (list of integers): Which days to repeat on, if any, 0=Monday)
        """
        timer_time = self._formatTimerTime(time, repeat_days)
        charge_timer = self._formatChargeTimer(charge, target_charge)
        full_timer = self._formatTimer(active, charge_timer, climate, timer_time)
        return await self._updateTimer(
            vin, full_timer, timerID, model, waitForConfirmation=waitForConfirmation
        )

    async def deleteTimer(self, vin, timerID="1", model=None, waitForConfirmation=True):
        """Delete an existing timer on the vehicle
        Parameters:
        vin (string): Vehicle VIN
        timerId (string): Target timer (numeric string)
        """
        return await self._deleteTimer(
            vin, timerID, model, waitForConfirmation=waitForConfirmation
        )

    async def lock(self, vin, waitForConfirmation=True):
        return await self._lockUnlock(
            vin, None, "quick-lock", waitForConfirmation=waitForConfirmation
        )

    async def unlock(self, vin, pin, waitForConfirmation=True):
        return await self._lockUnlock(
            vin, pin, "security-pin/unlock", waitForConfirmation=waitForConfirmation
        )

    async def climateOn(self, vin, waitForConfirmation=True):
        return await self._setClimate(
            vin, "true", waitForConfirmation=waitForConfirmation
        )

    async def climateOff(self, vin, waitForConfirmation=True):
        return await self._setClimate(
            vin, "false", waitForConfirmation=waitForConfirmation
        )

    async def directChargeOn(self, vin, model=None, waitForConfirmation=True):
        return await self._setDirectCharge(
            vin, "true", model, waitForConfirmation=waitForConfirmation
        )

    async def directChargeOff(self, vin, model=None, waitForConfirmation=True):
        return await self._setDirectCharge(
            vin, "false", model, waitForConfirmation=waitForConfirmation
        )

    async def honkAndFlash(self, vin, waitForConfirmation=True):
        return await self._setHonkAndFlash(vin, waitForConfirmation=waitForConfirmation)

    async def flash(self, vin, waitForConfirmation=True):
        return await self._setFlash(vin, waitForConfirmation=waitForConfirmation)

    async def getPermissions(self, vin):
        return await self._connection.get(
            f"https://api.porsche.com/core/api/v2/se/sv_SE/vehicles/{vin}/permissions"
        )

    async def isAllowed(self, vin):
        perms = await self.getPermissions(vin)
        print(perms)
        allowed = False
        reason = ""
        if perms["userIsActive"] and perms["userRoleStatus"] == "ENABLED":
            service_status = await self._connection.get(
                f"https://api.porsche.com/core/api/v4/de/de_DE/services?{vin}"
            )
            print(service_status)
            reason = service_status.get(
                "STATUS", service_status.get("disabledReason", "")
            )
            mydata = await self._connection.get(
                f"https://api.porsche.com/profiles/mydata?country={self.country}"
            )
            _LOGGER.debug(mydata)
            for vehicle in mydata["vehicles"]:
                print(vehicle)
                if vehicle["vin"] == vin:
                    if vehicle["confirmed"] and vehicle["pcc"]:
                        allowed = True
                        break
        return_value = {"allowed": allowed, "reason": reason}
        return return_value

    async def getVehicles(self):
        vehicles = await self._connection.get(
            f"https://api.porsche.com/core/api/v3/{self.locale_str}/vehicles"
        )
        return vehicles

    async def getStoredOverview(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/vehicle-data/{vin}/stored"
        )
        return data

    async def getCurrentOverview(self, vin):
        progressResult = await self._connection.post(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/vehicle-data/{vin}/current/request"
        )
        spinnerResult = await self._spinner(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/vehicle-data/{vin}/current/request/{progressResult['requestId']}/status"
        )
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/vehicle-data/{vin}/current/request/{progressResult['requestId']}"
        )
        return data

    async def getMaintenance(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/predictive-maintenance/information/{vin}"
        )
        return data

    async def getSummary(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/vehicle-summary/{vin}"
        )
        return data

    async def getServices(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/service-access/{vin}/details"
        )
        # Find out if we're in privacy mode
        is_privacy = data['vehicleServiceEnabledMap']['VSR'] == 'DISABLED'
        data["privacyMode"] = is_privacy
        return data

    async def getCapabilities(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/vcs/capabilities/{vin}"
        )
        return data

    async def getPosition(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/car-finder/{vin}/position"
        )
        return data

    async def getTheftAlerts(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/theft-alerts/{vin}/history"
        )
        return data

    async def getSpeedAlerts(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/speed-alert/{vin}/alerts"
        )
        return data

    async def getTripLongTerm(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/trips/{vin}/LONG_TERM/newest"
        )
        return data

    async def getTripShortTerm(self, vin):
        data = await self._connection.get(
            f"https://api.porsche.com/service-vehicle/{self.locale_str}/trips/{vin}/SHORT_TERM"
        )
        return data

    async def getEmobility(
        self,
        vin,
        model=None,
    ):
        if model is None:
            data = await self.getCapabilities(vin)
            model = data["carModel"]
        data = await self._connection.get(
            f"https://api.porsche.com/e-mobility/{self.locale_str}/{model}/{vin}?timezone={self.timezone}"
        )
        return data
